#!/usr/bin/env python
##
##  video.py - recoding VNC to FLV.
##
##  Copyright (c) 2009 by Yusuke Shinyama
##

import sys, zlib, re
from flvscreen import FlvScreen


##  Utilities
##
def str2clip(s):
  m = re.match(r'^(\d+)x(\d+)([\-\+])(\d+)([\-\+])(\d+)$', s)
  if not m:
    raise ValueError('Invalid clipping spec: %r' % s)
  return ((m.group(3), int(m.group(4))),
          (m.group(5), int(m.group(6))),
          int(m.group(1)), int(m.group(2)))

def str2size(s):
  m = re.match(r'^(\d+)x(\d+)$', s)
  if not m:
    raise ValueError('Invalid size spec: %r' % s)
  f = map(int, m.groups())
  return (f[0],f[1])


##  VideoSink
##
class VideoSink(object):

  def init_screen(self, width, height, name=None):
    print >>sys.stderr, 'init_screen: %dx%d, name=%r' % (width, height, name)
    return (0, 0, width, height)
  def init_clock(self, t):
    print >>sys.stderr, 'init_clock: t=%r' % (t)
    return

  # data is given as ARGB
  def convert_pixels(self, data):
    return data
  def convert_color1(self, data):
    return unpack('BBBx', data)

  def change_cursor(self, width, height, data):
    print >>sys.stderr, 'change_cursor: %dx%d' % (width, height)
    return

  def move_cursor(self, x, y):
    print >>sys.stderr, 'move_cursor: (%d,%d)' % (x,y)
    return

  def begin_update(self):
    print >>sys.stderr, 'begin_update'
    return

  def end_update(self, t):
    print >>sys.stderr, 'end_update', t
    return

  def idle(self, t):
    print >>sys.stderr, 'idle', t
    return
 
  def process_rgba(self, (x, y), (width, height), data):
    print >>sys.stderr, 'process_rgba: %dx%d at (%d,%d)' % (width,height,x,y)
    return
  
  def process_solid(self, (x, y), (w, h), data):
    print >>sys.stderr, 'process_solid: %dx%d at (%d,%d), color=%r' % (width,height,x,y, color)
    return

  def close(self, t):
    print >>sys.stderr, 'close', t
    return
  

##  FLVVideoSink
##
class FLVVideoSink(VideoSink):
  
  def __init__(self, writer, blocksize=32, framerate=15, keyframe=0,
               clipping=None, scrollwindow=None, scrollspeed=0, debug=0):
    self.debug = debug
    self.writer = writer
    self.blocksize = blocksize
    self.framerate = framerate
    self.keyframe = keyframe
    self.clipping = clipping
    self.scrollwindow = scrollwindow
    self.scrollspeed = scrollspeed
    self.t0 = None
    self.screen = None
    self.screenpos = (0,0)
    self.screensize = None
    self.windowpos = (0,0)
    self.windowsize = None
    self.t = 0
    self.curframe = 0
    self.centerpos = []
    return

  def init_screen(self, width, height, name=None):
    if self.clipping:
      ((xs,x), (ys,y), w, h) = self.clipping
      if xs == '-':
        (x,width) = (width-w-x,w)
      else:
        (x,width) = (x,w)
      if ys == '-':
        (y,height) = (height-h-x,h)
      else:
        (y,height) = (y,h)
    else:
      (x, y) = (0, 0)
    bw = (width+self.blocksize-1) / self.blocksize
    bh = (height+self.blocksize-1) / self.blocksize
    self.screenpos = (x,y)
    self.screensize = (bw,bh)
    self.screen = FlvScreen(self.blocksize, bw, bh)
    if self.scrollwindow:
      (w, h) = self.scrollwindow
      self.windowsize = ((w+self.blocksize-1) / self.blocksize,
                         (h+self.blocksize-1) / self.blocksize)
    else:
      self.windowsize = (bw, bh)
    if self.debug:
      print >>sys.stderr, 'start: %d,%d (%dx%d)' % (x, y, width, height)
    return (x, y, width, height)

  def init_clock(self, t):
    self.t0 = t
    return

  def change_cursor(self, width, height, data):
    return

  def move_cursor(self, x, y):
    return
 
  def process_rgba(self, (x, y), (w, h), data):
    (x0,y0) = self.screenpos
    self.screen.blit_rgba(x-x0, y-y0, w, h, data)
    return

  def process_solid(self, (x, y), (w, h), data):
    raise NotImplementedError

  def begin_update(self):
    return

  def end_update(self, t):
    self.idle(t)
    return

  def idle(self, t):
    self.t += t-self.t0
    while (self.curframe * 1000 / self.framerate) <= self.t:
      self.write_updates()
    self.t0 = t
    return

  def close(self, t):
    if self.screen:
      self.write_updates()
    return
  
  # write SCREENVIDEOPACKET tag
  def write_updates(self):
    changes = self.screen.changed()
    (bw,bh) = self.windowsize
    (bx,by) = self.do_autoscroll(self.windowpos, changes)
    key = ((bx,by) != self.windowpos or
           (self.keyframe and (self.curframe % self.keyframe) == 0))
    if key:
      # update the entire screen if necessary.
      self.windowpos = (bx,by)
      changes = set( (bx+x,by+y) for y in xrange(bh) for x in xrange(bw) )
    else:
      changes = set(changes)
    self.writer.start_tag()
    if self.debug:
      print >>sys.stderr, 'update(%d): changes=%r' % (self.curframe, len(changes)), sorted(changes)
    if key:
      self.writer.writebits(4, 1)
    else:
      self.writer.writebits(4, 2)
    self.writer.writebits(4, 3) # screenvideo codec
    self.writer.writebits(4, self.blocksize/16-1) # horizontal block size
    self.writer.writebits(12, bw*self.blocksize) # horizontal pixel width
    self.writer.writebits(4, self.blocksize/16-1) # vertical block size
    self.writer.writebits(12, bh*self.blocksize) # vertical pixel width
    self.writer.finishbits()
    n = 0
    for y in xrange(bh, 0, -1):
      y = by+y-1
      for x in xrange(bw):
        x += bx
        if (x,y) in changes:
          # changed block
          data = zlib.compress(self.screen.get(x,y))
          self.writer.writeub16(len(data))
          self.writer.write(data)
        else:
          # unchanged block
          self.writer.writeub16(0)
    t = (self.curframe * 1000 / self.framerate)
    self.writer.end_tag(9, t)
    self.screen.reset()
    self.curframe += 1
    return

  # do scrolling.
  def do_autoscroll(self, (wx,wy), changes):
    (ww,wh) = self.windowsize
    (bw,bh) = self.screensize
    (cx0,cy0) = (wx+ww/2, wy+wh/2)
    if changes:
      cx = sum( x for (x,y) in changes ) / len(changes)
      cy = sum( y for (x,y) in changes ) / len(changes)
    else:
      (cx,cy) = (cx0,cy0)
    self.centerpos = self.centerpos[len(self.centerpos)-self.scrollspeed:]+[(cx,cy)]
    cx1 = sum( x for (x,y) in self.centerpos ) / len(self.centerpos)
    cy1 = sum( y for (x,y) in self.centerpos ) / len(self.centerpos)
    if cx1 < cx0 and 0 < wx:
      wx -= 1
    elif cx0 < cx1 and wx+ww < bw-1:
      wx += 1
    if cy1 < cy0 and 0 < wy:
      wy -= 1
    elif cy0 < cy1 and wy+wh < bh-1:
      wy += 1
    return (wx,wy)


# main
if __name__ == '__main__':
  from vnc2flv.flv import FLVWriter
  from vnc2flv.rfb import RFBNetworkClient
  fp = file('out.flv', 'wb')
  writer = FLVWriter(fp)
  sink = FLVVideoSink(writer, debug=1)
  client = RFBNetworkClient('127.0.0.1', 5900, sink)
  client.open()
  try:
    while 1:
      client.idle()
  except KeyboardInterrupt:
    pass
  client.close()
  fp.close()
